from google.protobuf import message
from typing import Dict, List, Optional, Union
import collections
from typing import Any,  Dict, Iterable, List, Optional, Union
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf import any_pb2
from osmosis_proto.proto.osmosis.gamm.pool_models.stableswap import stableswap_pool_pb2
from osmosis_proto.proto.osmosis.gamm.pool_models.balancer import balancerPool_pb2

# number of decimal places
PRECISION = 18


def from_sdk_dec(dec_str: str) -> float:
    if dec_str is None or dec_str == '':
        return 0

    if '.' in dec_str:
        raise TypeError(
            f'expected a decimal string but got {dec_str} containing \'.\'')

    try:
        int(dec_str)
    except ValueError as err:
        raise ValueError(f'failed to convert {dec_str} to a number') from err

    neg = False
    if dec_str[0] == '-':
        neg = True
        dec_str = dec_str[1:]

    input_size = len(dec_str)
    bz_str = ''
    # case 1, purely decimal
    if input_size <= PRECISION:
        # 0. prefix
        bz_str = '0.'

        # set relevant digits to 0
        bz_str += '0' * (PRECISION - input_size)

        # set final digits
        bz_str += dec_str
    else:
        # inputSize + 1 to account for the decimal point that is being added
        dec_point_place = input_size - PRECISION

        bz_str = dec_str[:dec_point_place]  # pre-decimal digits
        bz_str += '.'  # decimal point
        bz_str += dec_str[dec_point_place:]  # pre-decimal digits

    if neg:
        bz_str = '-' + bz_str

    return float(bz_str)


def from_sdk_int(int_str: str) -> int:
    return int(int_str)


def dict_keys_must_match(dict_: dict, keys: List[str]):
    """Asserts that two iterables have the same elements, the same number of
    times, without regard to order.
    Alias for the 'element_counts_are_equal' function.

    dict_keys_must_match(dict_, keys)

    Example:
    - [0, 1, 1] and [1, 0, 1] compare equal.
    - [0, 0, 1] and [0, 1] compare unequal.

    """
    assert element_counts_are_equal(dict_.keys(), keys)


def element_counts_are_equal(
    first: Iterable[Any], second: Iterable[Any]
) -> Optional[bool]:
    """Asserts that two iterables have the same elements, the same number of
    times, without regard to order.

    Args:
        first (Iterable[Any])
        second (Iterable[Any])

    Returns:
        Optional[bool]: "passed" status. If this is True, first and second share
            the same element counts. If they don't the function will raise an
            AssertionError and return 'None'.
    """
    first_seq, second_seq = list(first), list(second)

    passed: Union[bool, None]
    try:
        first = collections.Counter(first_seq)
        second = collections.Counter(second_seq)
    except TypeError:
        # Handle case with unhashable elements
        differences = _count_diff_all_purpose(first_seq, second_seq)
    else:
        if first == second:
            passed = True
            return passed
        differences = _count_diff_hashable(first_seq, second_seq)

    if differences:
        standardMsg = "Element counts were not equal:\n"
        lines = ["First has %d, Second has %d:  %r" %
                 diff for diff in differences]
        diffMsg = "\n".join(lines)
        msg = "\n".join([standardMsg, diffMsg])
        passed = False
        assert passed, msg


_Mismatch = collections.namedtuple("Mismatch", "actual expected value")


def _count_diff_all_purpose(actual, expected):
    "Returns list of (cnt_act, cnt_exp, elem) triples where the counts differ"
    # elements need not be hashable
    s, t = list(actual), list(expected)
    m, n = len(s), len(t)
    NULL = object()
    result = []
    for i, elem in enumerate(s):
        if elem is NULL:
            continue
        cnt_s = cnt_t = 0
        for j in range(i, m):
            if s[j] == elem:
                cnt_s += 1
                s[j] = NULL
        for j, other_elem in enumerate(t):
            if other_elem == elem:
                cnt_t += 1
                t[j] = NULL
        if cnt_s != cnt_t:
            diff = _Mismatch(cnt_s, cnt_t, elem)
            result.append(diff)

    for i, elem in enumerate(t):
        if elem is NULL:
            continue
        cnt_t = 0
        for j in range(i, n):
            if t[j] == elem:
                cnt_t += 1
                t[j] = NULL
        diff = _Mismatch(0, cnt_t, elem)
        result.append(diff)
    return result


def _count_diff_hashable(actual, expected):
    "Returns list of (cnt_act, cnt_exp, elem) triples where the counts differ"
    # elements must be hashable
    s, t = collections.Counter(actual), collections.Counter(expected)
    result = []
    for elem, cnt_s in s.items():
        cnt_t = t.get(elem, 0)
        if cnt_s != cnt_t:
            diff = _Mismatch(cnt_s, cnt_t, elem)
            result.append(diff)
    for elem, cnt_t in t.items():
        if elem not in s:
            diff = _Mismatch(0, cnt_t, elem)
            result.append(diff)
    return result


PROTOBUF_MSG_BASE_ATTRS: List[str] = (
    dir(message.Message)
    + ['Extensions', 'FindInitializationErrors', '_CheckCalledFromGeneratedFile']
    + ['_extensions_by_name', '_extensions_by_number']
)
"""PROTOBUF_MSG_BASE_ATTRS (List[str]): The default attributes and methods of
an instance of the 'protobuf.message.Message' class.
"""


def camel_to_snake(camel: str):
    return ''.join(
        ['_' + char.lower() if char.isupper() else char for char in camel]
    ).lstrip('_')


def dict_keys_from_camel_to_snake(d):
    """
    Transform all keys from the dictionary from camelcase to snake case.

    Args:
        d (dict): The dictionary to transform

    Returns:
        dict: The dictionary transformed
    """
    if isinstance(d, list):
        return [
            dict_keys_from_camel_to_snake(
                i) if isinstance(i, (dict, list)) else i
            for i in d
        ]
    return {
        camel_to_snake(a): dict_keys_from_camel_to_snake(b)
        if isinstance(b, (dict, list))
        else b
        for a, b in d.items()
    }


def deserialize(pb_msg: message.Message, no_sdk_transformation: bool = False) -> dict:
    """Deserializes a proto message into a dictionary.

    - sdk.Dec values are converted to floats.
    - sdk.Int values are converted to ints.
    - Missing fields become blank strings.

    Args:
        pb_msg (protobuf.message.Message)
        no_sdk_transformation (bool): Wether to bypass the sdk transformation. Default to False

    Returns:
        dict: 'pb_msg' as a JSON-able dictionary.
    """

    if isinstance(pb_msg, any_pb2.Any):
        return pb_msg

    if not isinstance(pb_msg, message.Message):
        return pb_msg

    custom_dtypes: Dict[str, bytes] = {
        str(field[1]): field[0].GetOptions().__getstate__().get("serialized", None)
        for field in pb_msg.ListFields()
    }
    serialized_output = {}
    expected_fields: List[str] = list(pb_msg.DESCRIPTOR.fields_by_name.keys())

    for _, attr in enumerate(expected_fields):
        attr_search = pb_msg.__getattribute__(attr)
        custom_dtype = custom_dtypes.get(str(attr_search))

        if custom_dtype is not None:
            if "sdk/types.Dec" in str(custom_dtype):
                if no_sdk_transformation:
                    serialized_output[str(attr)] = float(
                        pb_msg.__getattribute__(attr))
                else:
                    serialized_output[str(attr)] = from_sdk_dec(
                        pb_msg.__getattribute__(attr)
                    )
            elif "sdk/types.Int" in str(custom_dtype):
                if no_sdk_transformation:
                    serialized_output[str(attr)] = int(
                        pb_msg.__getattribute__(attr))
                else:
                    serialized_output[str(attr)] = from_sdk_int(
                        pb_msg.__getattribute__(attr)
                    )
            elif "Int" in str(custom_dtype):  # Used for sdk.Coin message normalization
                serialized_output[str(attr)] = int(
                    pb_msg.__getattribute__(attr))
            else:
                val = pb_msg.__getattribute__(attr)
                if hasattr(val, '__len__') and not isinstance(val, str):
                    updated_vals = []
                    for v in val:
                        updated_vals.append(deserialize(v))
                    serialized_output[str(attr)] = updated_vals
                else:
                    serialized_output[str(attr)] = deserialize(val)
        elif custom_dtype is None and not attr_search:
            if str(attr_search) == "[]":
                serialized_output[str(attr)] = []
            else:
                serialized_output[str(attr)] = attr_search
        else:
            serialized_output[str(attr)] = deserialize(
                pb_msg.__getattribute__(attr))

    return serialized_output


def deserialize_pool(pool: any_pb2.Any):
    """
    Deserialize different pool types from Any type

    TODO: make this more generic and automate the process
    """
    if pool.Is(stableswap_pool_pb2.Pool.DESCRIPTOR):
        pool_obj = stableswap_pool_pb2.Pool()
    elif pool.Is(balancerPool_pb2.Pool.DESCRIPTOR):
        pool_obj = balancerPool_pb2.Pool()
    else:
        raise ValueError("Unknown pool type", pool.type_url)

    pool_obj.ParseFromString(pool.value)
    return deserialize(pool_obj)



"""
Classes:
    TxClient: A client for building, simulating, and broadcasting transactions.
    Transaction: Transactions trigger state changes based on messages. Each message
        must be cryptographically signed before being broadcasted to the network.
"""
import json
import logging
from copy import deepcopy
from numbers import Number
from typing import Any, Callable, Iterable, List, Tuple, Union

from google.protobuf import any_pb2, message
from google.protobuf.json_format import MessageToDict
from osmosis_proto.proto.cosmos.base.abci.v1beta1 import abci_pb2 as abci_type
from osmosis_proto.proto.cosmos.base.v1beta1 import coin_pb2 as cosmos_base_coin_pb
from osmosis_proto.proto.cosmos.base.v1beta1.coin_pb2 import Coin
from osmosis_proto.proto.cosmos.tx.signing.v1beta1 import signing_pb2 as tx_sign
from osmosis_proto.proto.cosmos.tx.v1beta1 import tx_pb2 as cosmos_tx_type

from osmosispy import pytypes as pt
from osmosispy import wallet
from osmosispy.exceptions import SimulationError, TxError
from osmosispy.grpc_client import GrpcClient


class TxClient:
    """
    The `TxClient` ia s class provided by the SDK to allow developers to build,
    simulate, and broadcast transactions on the blockchain.
    """

    def __init__(
        self,
        priv_key: wallet.PrivateKey,
        network: pt.Network,
        client: GrpcClient,
        config: pt.TxConfig,
    ):
        self.priv_key = priv_key
        self.network = network
        self.client = client
        self.address = None
        self.config = config

    def execute_tx(
        self, tx: "Transaction", gas_estimate: float, **kwargs
    ) -> abci_type.TxResponse:
        conf: pt.TxConfig = self.get_config(**kwargs)

        def compute_gas_wanted() -> float:
            # Related to https://github.com/cosmos/cosmos-sdk/issues/14405
            # TODO We should consider adding the behavior mentioned by tac0turtle.
            gas_wanted = gas_estimate * 1.25  # apply gas multiplier
            if conf.gas_wanted > 0:
                gas_wanted = conf.gas_wanted
            elif conf.gas_multiplier > 0:
                gas_wanted = gas_estimate * conf.gas_multiplier
            return gas_wanted

        gas_wanted = compute_gas_wanted()
        gas_price = pt.GAS_PRICE if conf.gas_price <= 0 else conf.gas_price

        fee = [
            cosmos_base_coin_pb.Coin(
                amount=str(int(gas_price * gas_wanted)),
            )
        ]
        logging.info(
            "Executing transaction with fee: %s and gas_wanted: %d", fee, gas_wanted
        )
        tx = (
            tx.with_gas_limit(gas_wanted)
            .with_fee(fee)
            .with_memo("")
            .with_timeout_height(self.client.timeout_height)
        )
        tx_raw_bytes = tx.get_signed_tx_data()

        return self._send_tx(tx_raw_bytes, conf.tx_type)

    def _send_tx(self, tx_raw_bytes: bytes, tx_type: pt.TxType) -> abci_type.TxResponse:
        broadcast_fn: Callable[[bytes], abci_type.TxResponse]

        if tx_type == pt.TxType.SYNC:
            broadcast_fn = self.client.send_tx_sync_mode
        elif tx_type == pt.TxType.ASYNC:
            broadcast_fn = self.client.send_tx_async_mode
        else:
            broadcast_fn = self.client.send_tx_block_mode

        return broadcast_fn(tx_raw_bytes)

    def build_tx(
        self,
        msgs: Union[pt.PythonMsg, List[pt.PythonMsg]],
        get_sequence_from_node: bool = False,
    ) -> Tuple["Transaction", wallet.Address]:
        if not isinstance(msgs, list):
            msgs = [msgs]

        pb_msgs = [msg.to_pb() for msg in msgs]

        self.client.sync_timeout_height()
        address: wallet.Address = self.get_address_info()
        sequence: int = address.get_sequence(
            from_node=get_sequence_from_node,
            lcd_endpoint=self.network.lcd_endpoint,
        )
        tx = (
            Transaction()
            .with_messages(pb_msgs)
            .with_sequence(sequence)
            .with_account_num(address.get_number())
            .with_chain_id(self.network.chain_id)
            .with_signer(self.priv_key)
        )
        return tx, address

    def simulate(self, tx: "Transaction") -> abci_type.SimulationResponse:
        """
        Simulate a transaction to estimate gas usage, validate the transaction parameters and determine if the transaction would be executed successfully.

        Args:
            tx (Transaction): The transaction object to be simulated.

        Returns:
            SimulationResponse: The simulation response, containing information about the simulated transaction.

        Raises:
            SimulationError: If there was an error while simulating the transaction.
        """
        sim_tx_raw_bytes: bytes = tx.get_signed_tx_data()

        sim_res: abci_type.SimulationResponse
        success: bool
        sim_res, success = self.client.simulate_tx(sim_tx_raw_bytes)
        if not success:
            raise SimulationError(sim_res)

        return sim_res

    def get_address_info(self) -> wallet.Address:
        """
        Get the address information of the account associated with the private key.

        Returns:
            Address: The address information of the account associated with the private key.
        """
        if self.address is None:
            pub_key: wallet.PublicKey = self.priv_key.to_public_key()
            self.address = pub_key.to_address()
            self.address = self.address.init_num_seq(self.network.lcd_endpoint)

        return self.address

    def get_config(self, **kwargs) -> pt.TxConfig:
        """
        Get the transaction configuration.

        Returns:
            TxConfig: The transaction configuration.
        """
        config: pt.TxConfig = deepcopy(self.config)
        prop_names = dir(config)
        for k, v in kwargs.items():
            if k in prop_names:
                setattr(config, k, v)
            else:
                logging.warning(
                    "%s is not a supported config property, ignoring", k)

        return config


class Transaction:
    """
    The state of a system changes when a `Transaction` is executed with one or
    more messages ('msgs'). In order for a message to be valid, it must be signed
    by an authorized account and then broadcasted to the network. Once the message
    is received by the network, it undergoes a validation process to ensure that
    it is secure and complies with the consensus rules of the network. Only after
    successful validation, it can be included in a block and added to the blockchain.

    Attributes:
        msgs (Tuple[Message]): The messages to be executed in the transaction.
        account_num (int): The number of the account in the state.
        sequence (int): Each transaction must have a unique sequence number
            that serves as a security measure to prevent replay attacks.
            The sequence number must be unique each time a transaction is
            broadcasted to the network.
        chain_id (str): The chain ID of the network.
        fee (List[Coin]): The fee to be paid for the transaction.
        gas_limit (int): The maximum amount of gas that can be used to execute
            the transaction.
        priv_key (wallet.PrivateKey): The private key of the account that will
            sign the transaction.
        memo (str): A note or description to be added to the transaction.
        timeout_height (int): The block height after which the transaction will
            be rejected.
    """

    def __init__(
        self,
        msgs: Tuple[message.Message, ...] = None,
        account_num: int = None,
        priv_key: wallet.PrivateKey = None,
        sequence: int = None,
        chain_id: str = None,
        fee: List[Coin] = None,
        gas_limit: int = 0,
        memo: str = "",
        timeout_height: int = 0,
    ):
        self.msgs = self.__convert_msgs(msgs) if msgs is not None else []
        self.account_num = account_num
        self.priv_key = priv_key
        self.sequence = sequence
        self.chain_id = chain_id
        self.fee = cosmos_tx_type.Fee(amount=fee, gas_limit=gas_limit)
        self.gas_limit = gas_limit
        self.memo = memo
        self.timeout_height = timeout_height

    @staticmethod
    def __convert_msgs(msgs: List[message.Message]) -> List[any_pb2.Any]:
        any_msgs: List[any_pb2.Any] = []
        for msg in msgs:
            any_msg = any_pb2.Any()
            any_msg.Pack(msg, type_url_prefix="")
            any_msgs.append(any_msg)
        return any_msgs

    def with_messages(self, msgs: Iterable[message.Message]) -> "Transaction":
        self.msgs.extend(self.__convert_msgs(msgs))
        return self

    def with_sender(self, client: GrpcClient, sender: str) -> "Transaction":
        if len(self.msgs) == 0:
            raise IndexError(
                "messsage is empty, please use with_messages at least 1 message"
            )
        account = client.get_account(sender)
        if account:
            self.account_num = account.account_number
            self.sequence = account.sequence
            return self
        raise KeyError("Account doesn't exist")

    def with_signer(self, priv_key: wallet.PrivateKey):
        self.priv_key = priv_key
        return self

    def with_account_num(self, account_num: int) -> "Transaction":
        self.account_num = account_num
        return self

    def with_sequence(self, sequence: int) -> "Transaction":
        self.sequence = sequence
        return self

    def with_chain_id(self, chain_id: str) -> "Transaction":
        self.chain_id = chain_id
        return self

    def with_fee(self, fee: List[Coin]) -> "Transaction":
        self.fee = cosmos_tx_type.Fee(amount=fee, gas_limit=self.fee.gas_limit)
        return self

    def with_gas_limit(self, gas: Number) -> "Transaction":
        self.fee.gas_limit = int(gas)
        return self

    def with_memo(self, memo: str) -> "Transaction":
        if len(memo) > pt.MAX_MEMO_CHARACTERS:
            raise ValueError("memo is too large")
        self.memo = memo
        return self

    def with_timeout_height(self, timeout_height: int) -> "Transaction":
        self.timeout_height = timeout_height
        return self

    def __generate_info(self, public_key: wallet.PublicKey = None) -> Tuple[str, str]:
        body = cosmos_tx_type.TxBody(
            messages=self.msgs, memo=self.memo, timeout_height=self.timeout_height
        )

        body_bytes = body.SerializeToString()
        mode_info = cosmos_tx_type.ModeInfo(
            single=cosmos_tx_type.ModeInfo.Single(
                mode=tx_sign.SIGN_MODE_DIRECT)
        )

        if public_key:
            any_public_key = any_pb2.Any()
            any_public_key.Pack(
                public_key.to_public_key_proto(), type_url_prefix="")
            signer_info = cosmos_tx_type.SignerInfo(
                mode_info=mode_info, sequence=self.sequence, public_key=any_public_key
            )
        else:
            signer_info = cosmos_tx_type.SignerInfo(
                mode_info=mode_info, sequence=self.sequence
            )

        auth_info = cosmos_tx_type.AuthInfo(
            signer_infos=[signer_info], fee=self.fee)
        auth_info_bytes = auth_info.SerializeToString()

        return body_bytes, auth_info_bytes

    def get_sign_doc(
        self, public_key: wallet.PublicKey = None
    ) -> cosmos_tx_type.SignDoc:
        if len(self.msgs) == 0:
            raise ValueError("message is empty")

        if self.account_num is None:
            raise RuntimeError("account_num should be defined")

        if self.sequence is None:
            raise RuntimeError("sequence should be defined")

        if self.chain_id is None:
            raise RuntimeError("chain_id should be defined")

        body_bytes, auth_info_bytes = self.__generate_info(public_key)

        return cosmos_tx_type.SignDoc(
            body_bytes=body_bytes,
            auth_info_bytes=auth_info_bytes,
            chain_id=self.chain_id,
            account_number=self.account_num,
        )

    def get_tx_data(
        self, signature: bytes, public_key: wallet.PublicKey = None
    ) -> bytes:
        body_bytes, auth_info_bytes = self.__generate_info(public_key)

        tx_raw = cosmos_tx_type.TxRaw(
            body_bytes=body_bytes,
            auth_info_bytes=auth_info_bytes,
            signatures=[signature],
        )
        return tx_raw.SerializeToString()

    def get_signed_tx_data(self) -> bytes:
        if self.priv_key is None:
            raise RuntimeError("priv_key should be defined")

        pub_key = self.priv_key.to_public_key()
        sign_doc = self.get_sign_doc(pub_key)
        sig = self.priv_key.sign(sign_doc.SerializeToString())
        return self.get_tx_data(sig, pub_key)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.any_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class GenesisState(google.protobuf.message.Message):
    """GenesisState defines the evidence module's genesis state."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EVIDENCE_FIELD_NUMBER: builtins.int
    @property
    def evidence(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[google.protobuf.any_pb2.Any]:
        """evidence defines all the evidence at genesis."""
    def __init__(
        self,
        *,
        evidence: collections.abc.Iterable[google.protobuf.any_pb2.Any] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["evidence", b"evidence"]) -> None: ...

global___GenesisState = GenesisState



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cosmos/base/snapshots/v1beta1/snapshot.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gogoproto import gogo_pb2 as gogoproto_dot_gogo__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,cosmos/base/snapshots/v1beta1/snapshot.proto\x12\x1d\x63osmos.base.snapshots.v1beta1\x1a\x14gogoproto/gogo.proto\"\x89\x01\n\x08Snapshot\x12\x0e\n\x06height\x18\x01 \x01(\x04\x12\x0e\n\x06\x66ormat\x18\x02 \x01(\r\x12\x0e\n\x06\x63hunks\x18\x03 \x01(\r\x12\x0c\n\x04hash\x18\x04 \x01(\x0c\x12?\n\x08metadata\x18\x05 \x01(\x0b\x32\'.cosmos.base.snapshots.v1beta1.MetadataB\x04\xc8\xde\x1f\x00\" \n\x08Metadata\x12\x14\n\x0c\x63hunk_hashes\x18\x01 \x03(\x0c\"\x8f\x03\n\x0cSnapshotItem\x12\x41\n\x05store\x18\x01 \x01(\x0b\x32\x30.cosmos.base.snapshots.v1beta1.SnapshotStoreItemH\x00\x12I\n\x04iavl\x18\x02 \x01(\x0b\x32/.cosmos.base.snapshots.v1beta1.SnapshotIAVLItemB\x08\xe2\xde\x1f\x04IAVLH\x00\x12I\n\textension\x18\x03 \x01(\x0b\x32\x34.cosmos.base.snapshots.v1beta1.SnapshotExtensionMetaH\x00\x12T\n\x11\x65xtension_payload\x18\x04 \x01(\x0b\x32\x37.cosmos.base.snapshots.v1beta1.SnapshotExtensionPayloadH\x00\x12H\n\x0b\x61pp_version\x18\x05 \x01(\x0b\x32\x31.cosmos.base.snapshots.v1beta1.SnapshotAppVersionH\x00\x42\x06\n\x04item\"!\n\x11SnapshotStoreItem\x12\x0c\n\x04name\x18\x01 \x01(\t\"O\n\x10SnapshotIAVLItem\x12\x0b\n\x03key\x18\x01 \x01(\x0c\x12\r\n\x05value\x18\x02 \x01(\x0c\x12\x0f\n\x07version\x18\x03 \x01(\x03\x12\x0e\n\x06height\x18\x04 \x01(\x05\"5\n\x15SnapshotExtensionMeta\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06\x66ormat\x18\x02 \x01(\r\"+\n\x18SnapshotExtensionPayload\x12\x0f\n\x07payload\x18\x01 \x01(\x0c\"%\n\x12SnapshotAppVersion\x12\x0f\n\x07version\x18\x01 \x01(\x04\x42.Z,github.com/cosmos/cosmos-sdk/snapshots/typesb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'cosmos.base.snapshots.v1beta1.snapshot_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z,github.com/cosmos/cosmos-sdk/snapshots/types'
  _SNAPSHOT.fields_by_name['metadata']._options = None
  _SNAPSHOT.fields_by_name['metadata']._serialized_options = b'\310\336\037\000'
  _SNAPSHOTITEM.fields_by_name['iavl']._options = None
  _SNAPSHOTITEM.fields_by_name['iavl']._serialized_options = b'\342\336\037\004IAVL'
  _SNAPSHOT._serialized_start=102
  _SNAPSHOT._serialized_end=239
  _METADATA._serialized_start=241
  _METADATA._serialized_end=273
  _SNAPSHOTITEM._serialized_start=276
  _SNAPSHOTITEM._serialized_end=675
  _SNAPSHOTSTOREITEM._serialized_start=677
  _SNAPSHOTSTOREITEM._serialized_end=710
  _SNAPSHOTIAVLITEM._serialized_start=712
  _SNAPSHOTIAVLITEM._serialized_end=791
  _SNAPSHOTEXTENSIONMETA._serialized_start=793
  _SNAPSHOTEXTENSIONMETA._serialized_end=846
  _SNAPSHOTEXTENSIONPAYLOAD._serialized_start=848
  _SNAPSHOTEXTENSIONPAYLOAD._serialized_end=891
  _SNAPSHOTAPPVERSION._serialized_start=893
  _SNAPSHOTAPPVERSION._serialized_end=930
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import cosmos.mint.v1beta1.mint_pb2
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class QueryParamsRequest(google.protobuf.message.Message):
    """QueryParamsRequest is the request type for the Query/Params RPC method."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___QueryParamsRequest = QueryParamsRequest

@typing_extensions.final
class QueryParamsResponse(google.protobuf.message.Message):
    """QueryParamsResponse is the response type for the Query/Params RPC method."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PARAMS_FIELD_NUMBER: builtins.int
    @property
    def params(self) -> cosmos.mint.v1beta1.mint_pb2.Params:
        """params defines the parameters of the module."""
    def __init__(
        self,
        *,
        params: cosmos.mint.v1beta1.mint_pb2.Params | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["params", b"params"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["params", b"params"]) -> None: ...

global___QueryParamsResponse = QueryParamsResponse

@typing_extensions.final
class QueryInflationRequest(google.protobuf.message.Message):
    """QueryInflationRequest is the request type for the Query/Inflation RPC method."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___QueryInflationRequest = QueryInflationRequest

@typing_extensions.final
class QueryInflationResponse(google.protobuf.message.Message):
    """QueryInflationResponse is the response type for the Query/Inflation RPC
    method.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INFLATION_FIELD_NUMBER: builtins.int
    inflation: builtins.bytes
    """inflation is the current minting inflation value."""
    def __init__(
        self,
        *,
        inflation: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["inflation", b"inflation"]) -> None: ...

global___QueryInflationResponse = QueryInflationResponse

@typing_extensions.final
class QueryAnnualProvisionsRequest(google.protobuf.message.Message):
    """QueryAnnualProvisionsRequest is the request type for the
    Query/AnnualProvisions RPC method.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___QueryAnnualProvisionsRequest = QueryAnnualProvisionsRequest

@typing_extensions.final
class QueryAnnualProvisionsResponse(google.protobuf.message.Message):
    """QueryAnnualProvisionsResponse is the response type for the
    Query/AnnualProvisions RPC method.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ANNUAL_PROVISIONS_FIELD_NUMBER: builtins.int
    annual_provisions: builtins.bytes
    """annual_provisions is the current minting annual provisions value."""
    def __init__(
        self,
        *,
        annual_provisions: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["annual_provisions", b"annual_provisions"]) -> None: ...

global___QueryAnnualProvisionsResponse = QueryAnnualProvisionsResponse



from typing import Callable, Optional
from .client import ITradingClient
from .trade_data import TradeData
from .types import TradingFee
from osmosispy import Coin, utils, Sdk

from osmosis_proto.proto.osmosis.txfees.v1beta1 import (
    query_pb2 as txfees_query,
    query_pb2_grpc as txfees_query_grpc,
)


class OsmosisTradingClient(ITradingClient):
    trader: Sdk

    def __init__(self, trader: Sdk):
        self.trader = trader

    def get_tx_fee(self, symbol: str = "OSMOUSDT") -> Optional[Coin]:
        raise NotImplementedError

    def get_trading_fee(self, symbol: str = "OSMOUSDT") -> Optional[TradingFee]:
        raise NotImplementedError

    def get_price(self, symbol: str = "OSMOUSDT") -> Coin:
        raise NotImplementedError

    def listen_trades(self, cb: Callable[[TradeData], None], symbol: str = "OSMOUSDT"):
        raise NotImplementedError

    def stop_listeners(self) -> None:
        raise NotImplementedError



from typing import Callable, Optional
from os import getenv
from binance.websocket.spot.websocket_client import SpotWebsocketClient
from binance.spot import Spot

from .trade_data import TradeData
from .types import TradingFee
from .client import ITradingClient
from osmosispy import Coin


class BinanceTradingClient(ITradingClient):
    """
    Binance trading client
    """

    ws_client: SpotWebsocketClient
    api_client: Spot
    last_id = 0

    def __init__(
        self,
        api_key: str,
        api_secret: str,
    ):
        self.api_client = Spot(
            api_key=api_key,
            api_secret=api_secret,
        )

        self.ws_client = SpotWebsocketClient()

    @classmethod
    def _get_env_api_key(cls) -> str:
        res = getenv("BINANCE_API_KEY")

        if not res:
            raise ValueError("BINANCE_API_KEY not set")

        return res

    @classmethod
    def _get_env_api_secret(cls) -> str:
        res = getenv("BINANCE_API_SECRET")

        if not res:
            raise ValueError("BINANCE_API_SECRET not set")

        return res

    @classmethod
    def from_env(cls) -> "BinanceTradingClient":
        """
        Provides a convenient way to create a `TradingClient` instance without having to pass
        the api key and secret manually.

        Make sure to set the `BINANCE_API_KEY` and `BINANCE_API_SECRET` environment variables.

        ```bash
        export BINANCE_API_KEY=your_api_key
        export BINANCE_API_SECRET=your_api_secret
        ```

        Returns:
            TradingClient: The `TradingClient` instance
        """
        return cls(
            api_key=cls._get_env_api_key(),
            api_secret=cls._get_env_api_secret(),
        )

    def get_trading_fee(self, symbol: str = "OSMOUSDT") -> Optional[TradingFee]:
        res = self.api_client.trade_fee(symbol=symbol)
        try:
            fees = res[0]
            return TradingFee(
                maker_commision=float(fees["makerCommission"]),
                taker_commision=float(fees["takerCommission"]),
            )
        except (NameError, AttributeError, IndexError):
            return None

    def get_price(self, symbol: str = "OSMOUSDT") -> Coin:
        res = self.api_client.ticker_price(symbol=symbol)

        print(res)

        return Coin(denom=self.split_symbol(symbol)[1], amount=res["price"])

    def listen_trades(self, cb: Callable[[TradeData], None], symbol: str = "OSMOUSDT"):
        self.ws_client.start()

        id = self.last_id
        self.last_id += 1

        def callback(data):
            # check if the event is an aggTrade
            try:
                if data["e"] != "aggTrade":
                    return
            except KeyError as e:
                return

            quantity_asset, price_asset = self.split_symbol(symbol)

            # call the actual callback
            cb(TradeData(
                price=Coin(denom=price_asset, amount=data["p"]),
                quantity=Coin(denom=quantity_asset, amount=data["q"]),
            ))

        self.ws_client.agg_trade(
            id=id,
            symbol=symbol,
            callback=callback,
        )

    def stop_listeners(self) -> None:
        self.ws_client.stop()



"""
The fundamental gateway to the blockchain is the "Sdk" - an essential library that requires authorization
with a wallet/signer for each "Sdk" object.

The "Sdk" provides two crucial clients - a transaction client for signing and broadcasting transactions,
and a query client for sending gRPC queries. The functionalities offered by the "Sdk" are highly dependent
on the user's network and transaction configurations.

To configure the "Sdk" object, users can utilize the Network and TxConfig classes provided by the
osmosispy/pytypes package. These classes ensure seamless integration and optimized performance with
the desired network and transaction settings.
"""
import logging
from .pytypes import Network, TxConfig
from .grpc_client import GrpcClient
from .tx import TxClient
from .wallet import PrivateKey


class Sdk:
    """
    The Sdk class acts as an intermediary that facilitates the signing and sending of transactions or
    query execution from a node.

    It is associated with several crucial components, including a wallet or signer that can either be
    newly generated or recovered from an existing mnemonic. Additionally, the network that the node will
    connect to and an optional configuration defining how to behave and the gas configuration for each
    transaction are specified.

    Any method that starts with with_ will replace the existing Sdk object with a new version that has
    the defined behavior. This approach ensures flexibility and allows users to customize the Sdk object
    to suit their specific requirements.

    The Sdk class has the following essential attributes: priv_key, query, tx, network, and tx_config.

    Here's an example of how to use the Sdk class to create a new object with specific configurations:

    ```python
    sdk = (
        Sdk.authorize(val_mnemonic)
        .with_config(tx_config)
        .with_network(network)
    )
    ```

    By leveraging this code snippet, you can authorize a new Sdk object,
    configure it with a specific tx_config and network,
    and use it to execute your desired transactions or queries.
    """

    query: GrpcClient
    network: Network
    tx: TxClient
    tx_config: TxConfig
    mnemonic: str

    def __init__(self, _error_do_not_use_init_directly=None) -> None:
        """Unsupported, please use from_mnemonic to initialize."""
        if not _error_do_not_use_init_directly:
            raise TypeError(
                "Please use PrivateKey.from_mnemonic() to construct me")
        self.priv_key: PrivateKey = None
        self.query = None
        self.tx = None
        self.network = None
        self.tx_config = TxConfig()

    @classmethod
    def authorize(cls, key: str = None) -> "Sdk":
        """
        The Authorize method enables users to generate or recover a wallet and register it as an Sdk object.
        In the case of wallet recovery, users need to provide the mnemonic phrase via the key argument
        If the key argument is not provided, a new wallet is created automatically.

        Args:
            key (str, optional): The mnemonic phrase to recover the wallet. Ceating a new wallet if not provided.

        Returns:
            Sdk: The updated sdk object, which includes the registered wallet.

        By leveraging this method, users can quickly and efficiently generate or recover a wallet and register
        it as an Sdk object.
        """
        self = cls(_error_do_not_use_init_directly=True)
        if key is None:
            (mnemonic, pk) = PrivateKey.generate()
            logging.warning(
                "The mnemonic used for the newly generated key is: \n%s", mnemonic
            )
            logging.warning(
                "Please write down this key, it will NOT be recoverable otherwise"
            )
        elif len(key.split(" ")) > 1:
            pk = PrivateKey.from_mnemonic(key)
        elif len(key) > 0:
            pk = PrivateKey.from_hex(key)

        self._with_priv_key(pk)
        self.mnemonic = mnemonic if key is None else key

        return self

    def with_network(
        self, network: Network
    ) -> "Sdk":
        """
        `with_network` method allows users to modify the network configuration of the Sdk object to connect to a specified network.

        Args:
            network (Network): A network object that specifies the desired network configuration.

        Returns:
            Sdk: The updated sdk object with the new network configuration.

        By utilizing this method, users can seamlessly switch between different networks and enjoy optimal performance and security.
        """
        self.network = network
        self._with_query_client(
            GrpcClient(network, True)
        )
        return self

    def with_config(self, config: TxConfig) -> "Sdk":
        """
        Change the configuration for trasnactions for the sdk to the specified config.

        Args:
            config (TxConfig): A transaction configuration object

        Returns:
            Sdk: The updated sdk object
        """
        self.tx_config = config
        tx_client = TxClient(
            client=self.query,
            network=self.network,
            priv_key=self.priv_key,
            config=self.tx_config,
        )
        self._with_tx_client(tx_client)
        return self

    @property
    def address(self) -> str:
        """
        Public address of the wallet.

        Returns:
            str: The public address of the wallet
        """
        pub_key = self.priv_key.to_public_key()
        return pub_key.to_address().to_acc_bech32()

    # Private methods
    def _with_query_client(self, client: GrpcClient) -> "Sdk":
        self.query = client
        tx_client = TxClient(
            client=self.query,
            network=self.network,
            priv_key=self.priv_key,
            config=self.tx_config,
        )
        self._with_tx_client(tx_client)
        return self

    def _with_tx_client(self, tx_client: TxClient) -> "Sdk":
        self.tx = tx_client
        return self

    def _with_priv_key(self, priv_key: PrivateKey) -> "Sdk":
        self.priv_key = priv_key
        return self



from typing import Callable, Optional
from .trade_data import TradeData
from .types import TradingFee
from osmosispy import Coin
import re


class ITradingClient:
    """
    Interface for trading clients
    """

    @classmethod
    def split_symbol(cls, symbol: str) -> tuple[str, str]:
        """
        Splits the symbol into the two assets

        Args:
            symbol (str): The symbol to split

        Returns:
            tuple[str, str]: The two assets
        """

        v = re.split(r"([A-Z]+)(BTC|ETH|USDT|BNB|PAX|USDC|USDS)", symbol)

        return v[1], v[2]

    def get_trading_fee(self, symbol: str = "OSMOUSDT") -> Optional[TradingFee]:
        """
        Get the current trading fee for the symbol

        Args:
            symbol (str, optional): The pair to get the trading fee for. Defaults to "OSMOUSDT".

        Returns:
            Optional[TradingFee]: The trading fee for the pair. None if the there is no trades or fee info for the pair.
        """
        raise NotImplementedError

    def get_price(self, symbol: str = "OSMOUSDT") -> Coin:
        """
        Get the current price in the pair

        Args:
            symbol (str, optional): The pair to get the price in. Defaults to "OSMOUSDT".

        Returns:
            Coin: The price of the first asset in the pair
        """
        raise NotImplementedError

    def listen_trades(self, cb: Callable[[TradeData], None], symbol: str = "OSMOUSDT"):
        """
        Listen to trades for the pair

        Args:
            cb (`TradeData`) -> None: The callback to call when a trade is made
            symbol (str, optional): The pair to listen to. Defaults to "OSMOUSDT".

        Example:
        ```
        # Listen to trades for 60 seconds
        def cb(trade_data: TradeData):
            print(trade_data)

        client.listen_trades(cb)

        sleep(60)

        client.stop_ws()
        """
        raise NotImplementedError

    def stop_listeners(self) -> None:
        """
        Stops all listeners
        """
        raise NotImplementedError
